from wordfreq.config import DB_FILENAME, CACHE_SIZE
from wordfreq.util import standardize_word
import sqlite3
import sys

if sys.version_info.major == 2:
    from functools32 import lru_cache
else:
    from functools import lru_cache

SQLITE_ERROR_TEXT = """
Couldn't open the wordlist database.
You may need to run wordfreq's setup.py script.

I was expecting to find the database at:

    %(path)s

This can be configured by setting the WORDFREQ_DATA environment variable.
""" % {'path': DB_FILENAME}

try:
    CONN = sqlite3.connect(DB_FILENAME)
except sqlite3.OperationalError:
    raise IOError(SQLITE_ERROR_TEXT)


@lru_cache(maxsize=CACHE_SIZE)
def word_frequency(word, lang, wordlist='multi', offset=0.):
    """
    Get the frequency of `word` in the language with code `lang`, from the
    specified `wordlist`.

    The offset gets added to all values, to monotonically account for the
    fact that we have not observed all possible words.
    """
    c = CONN.cursor()
    c.execute("SELECT freq from words where word=? and lang=? and wordlist=?",
              (standardize_word(word), lang, wordlist))
    row = c.fetchone()
    if row is None:
        return offset
    else:
        return row[0] + offset


def wordlist_size(wordlist, lang=None):
    """
    Get the number of words in a wordlist.
    """
    c = CONN.cursor()
    if lang is None:
        c.execute(
            "SELECT count(*) from words where wordlist=?",
            (wordlist,)
        )
    else:
        c.execute(
            "SELECT count(*) from words where wordlist=? and lang=?",
            (wordlist, lang)
        )
    return c.fetchone()[0]


def average_frequency(wordlist, lang):
    """
    A kind of slow function to get the average frequency for words in a
    wordlist.

    If, for example, you're smoothing over word frequencies by adding the
    same baseline number to all of them, this can tell you what a good
    baseline is. (For multi/en, it's 6.7e-07.)
    """
    c = CONN.cursor()
    c.execute("SELECT avg(freq) from words where wordlist=? and lang=?",
              (wordlist, lang))
    return c.fetchone()[0]


def iter_wordlist(wordlist='multi', lang=None):
    """
    Returns a generator, yielding (word, lang, frequency) triples from
    a wordlist in descending order of frequency.

    If a `lang` is specified, the results will only contain words in that
    language.
    """
    c = CONN.cursor()
    if lang is None:
        results = c.execute(
            "SELECT word, lang, freq from words where wordlist=? "
            "ORDER BY freq desc",
            (wordlist,)
        )
    else:
        results = c.execute(
            "SELECT word, lang, freq from words where "
            "wordlist=? and lang=? ORDER BY freq DESC",
            (wordlist, lang)
        )

    return results


def wordlist_info(connection=None):
    """
    Get info about all the wordlists in a database, returning their
    list name, language, and number of words as 'wordlist', 'lang',
    and 'count' respectively.

    The database connection can be given as an argument, in order to get
    information about a database other than the default configured one.
    """
    if connection is None:
        connection = CONN
    c = connection.cursor()
    results = c.execute(
        "SELECT wordlist, lang, count(*) from words GROUP BY wordlist, lang"
    )
    for wordlist, lang, count in results:
        yield {'wordlist': wordlist, 'lang': lang, 'count': count}


def random_words(nwords=4, bits_per_word=12, wordlist='google-books',
                 lang='en'):
    """
    There are a few reasons you might want to see a sample of words in a
    wordlist:

    - Generating test cases
    - Getting a feel for what a wordlist contains
    - Generating passwords as in https://xkcd.com/936/

    Parameters:

    - `nwords` is the number of words to select.
    - `bits_per_word` indicate how many bits of randomness per word you want,
      up to log2(wordlist_size). As you increase it, the words get obscure.
    - `wordlist` and `lang` specify the wordlist to use.
    """
    import random
    limit = 2 ** bits_per_word
    c = CONN.cursor()
    results = c.execute(
        "SELECT word from words where wordlist = ? and lang = ? "
        "ORDER BY freq DESC LIMIT ?",
        (wordlist, lang, limit)
    )
    words = [row[0] for row in results]
    selected = random.sample(words, nwords)
    return u' '.join(selected)


def wordy_hash(item, nwords=2, bits_per_word=16, wordlist='google-books', lang='en'):
    limit = 2 ** bits_per_word
    c = CONN.cursor()
    results = c.execute(
        "SELECT word from words where wordlist = ? and lang = ? "
        "ORDER BY freq DESC LIMIT ?",
        (wordlist, lang, limit)
    )
    words = [row[0] for row in results]
    out_words = []
    value = abs(hash(item))
    while value and len(out_words) < nwords:
        out_words.append(words[value % limit])
        value /= limit
    out_words.reverse()
    return u' '.join(out_words)



SCHEMA = """
CREATE TABLE IF NOT EXISTS words (
    wordlist text,
    lang text,
    word text,
    freq real
)
"""

INDICES = [
    "CREATE UNIQUE INDEX IF NOT EXISTS words_uniq ON words (wordlist, lang, word)",
    "CREATE INDEX IF NOT EXISTS words_by_freq ON words (wordlist, lang, freq DESC)",
]



from collections import defaultdict
import sqlite3
import codecs
import re
import os
import sys
import logging
logger = logging.getLogger(__name__)

from ftfy import ftfy
from wordfreq import config, schema
from wordfreq.util import standardize_word


def read_csv(filename):
    """
    Load word frequencies from a file of comma-separated values, where
    each line contains a word, a comma, and its frequency.

    Scale the frequencies so they add up to 1.0, and return them as a
    dictionary.
    """
    return _scale_freqs(_read_csv_basic(filename))


def read_multilingual_csv(filename):
    """
    Load word frequencies from a file of comma-separated values, where
    each line is of the form:

        word|lang,freq

    Scale the frequencies so they add up to 1.0 *for each language*,
    and return a dictionary from language -> (word -> freq).
    """
    unscaled = defaultdict(dict)
    raw_freqs = _read_csv_basic(filename)
    for wordlang in raw_freqs:
        word, lang = wordlang.rsplit('|', 1)
        word = standardize_word(word)
        unscaled[lang][word] = raw_freqs[wordlang]

    scaled = {}
    for key in unscaled:
        scaled[key] = _scale_freqs(unscaled[key])
    return scaled


def _read_csv_basic(filename):
    infile = codecs.open(filename, encoding='utf-8')

    counts = {}
    for line in infile:
        line = line.rstrip(u'\n')
        word, count = line.rsplit(u',', 1)
        count = float(count)
        counts[standardize_word(word)] = count
    return counts


NUMBER_RE = re.compile(u'[0-9]+')
def read_leeds_corpus(filename):
    """
    Load word frequencies from a "Web as Corpus" file, collected and
    provided by the University of Leeds.

    For more information, see: http://corpus.leeds.ac.uk/list.html
    """
    infile = codecs.open(filename, encoding='utf-8')

    counts = defaultdict(float)
    for line in infile:
        line = line.rstrip()
        if line:
            rank = line.split(u' ')[0]
            if NUMBER_RE.match(rank) and line.count(u' ') == 2:
                _, freq, token = line.split(u' ')
                token = standardize_word(ftfy(token))
                freq = float(freq)
                counts[token] += freq

    return _scale_freqs(counts)


def _scale_freqs(counts):
    """
    Take in unscaled word counts or frequencies, and scale them so that
    they add up to 1.0.
    """
    freqs = {}
    total = sum(counts.values())
    for word in counts:
        freqs[word] = counts[word] / total

    return freqs


def save_wordlist_to_db(conn, listname, lang, freqs):
    """
    Save a dictionary of word frequencies to a database.

    The dictionary `freqs` should be properly scaled (run it through
    `_scale_freqs`). It will be saved as language `lang` in wordlist
    `listname`.
    """
    rows = [(listname, lang, word, freq)
            for word, freq in freqs.items()]
    conn.execute('DELETE FROM words where wordlist=? and lang=?',
                 (listname, lang))
    conn.executemany(
        "INSERT INTO words (wordlist, lang, word, freq) "
        "VALUES (?, ?, ?, ?)",
        rows
    )
    conn.commit()


def create_db(filename):
    """
    Create a wordlist database, at the filename specified by `wordfreq.config`.

    This should be safe to run (and have no effect) if the database already
    exists.
    """
    conn = get_db_connection(filename)
    base_dir = os.path.dirname(filename)
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)

    conn.execute(schema.SCHEMA)
    for index_definition in schema.INDICES:
        conn.execute(index_definition)
    conn.commit()


def get_db_connection(filename):
    return sqlite3.connect(filename)


LEEDS_LANGUAGES = ('ar', 'de', 'el', 'es', 'fr', 'it', 'ja', 'pt', 'ru', 'zh')
def load_all_data(source_dir=None, filename=None, do_it_anyway=False):
    """
    Load data from the raw data files into a SQLite database.

    Python 3 has more complete Unicode support than Python 2, and this shows
    up as actual differences in the set of words. For the sake of consistency,
    we say that the data is only valid when built on Python 3.

    Python 2 can still *use* wordfreq, by downloading the database that was
    built on Python 3.

    If you insist on building the Python 2 version, pass `do_it_anyway=True`.
    """
    if sys.version_info.major == 2 and not do_it_anyway:
        raise UnicodeError(
            "Python 2.x has insufficient Unicode support, and will build "
            "the wrong database. Pass `do_it_anyway=True` to do it anyway."
        )

    if source_dir is None:
        source_dir = config.RAW_DATA_DIR

    if filename is None:
        filename = config.DB_FILENAME

    logger.info("Creating database")
    create_db(filename)

    conn = get_db_connection(filename)
    logger.info("Loading Leeds internet corpus:")
    for lang in LEEDS_LANGUAGES:
        logger.info("\tLanguage: %s" % lang)
        filename = os.path.join(
            source_dir, 'leeds', 'internet-%s-forms.num' % lang
        )
        wordlist = read_leeds_corpus(filename)
        save_wordlist_to_db(conn, 'leeds-internet', lang, wordlist)

    logger.info("Loading Google Books (English).")
    google_wordlist = read_csv(
        os.path.join(source_dir, 'google', 'google-books-english.csv')
    )
    save_wordlist_to_db(conn, 'google-books', 'en', google_wordlist)

    logger.info("Loading combined multilingual corpus:")
    multi_wordlist = read_multilingual_csv(
        os.path.join(source_dir, 'luminoso', 'multilingual.csv')
    )
    for lang in multi_wordlist:
        logger.info("\tLanguage: %s" % lang)
        save_wordlist_to_db(conn, 'multi', lang, multi_wordlist[lang])

    logger.info("Loading Twitter corpus.")
    twitter_wordlist = read_csv(
        os.path.join(source_dir, 'luminoso', 'twitter-52M.csv')
    )
    save_wordlist_to_db(conn, 'twitter', 'xx', twitter_wordlist)

    logger.info("Done loading.")


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    load_all_data()



"""
This module provides some functions that help us work with the
wordlist data files, so that they don't have to be stored in Git
or in the source package.

These functions won't be used in the course of using the wordfreq
package normally; instead, they're called from commands in setup.py.
"""

from wordfreq import config
import os
import sys
import shutil
import tempfile
import tarfile
import logging
import subprocess
logger = logging.getLogger(__name__)

if sys.version_info.major == 2:
    from urllib import urlretrieve
else:
    from urllib.request import urlretrieve


class ProgressTracker(object):
    """
    This class watches the progress of a urllib download task, and updates
    sys.stdout when the percentage changes.
    """
    def __init__(self, url):
        self.url = url
        self.progress = None

    def report_progress(self, count, blockSize, totalSize):
        percent = int(count*blockSize*100/totalSize)
        if percent != self.progress:
            sys.stdout.write("\rDownloading %s... %2d%%" % (self.url, percent))
            sys.stdout.flush()
            self.progress = percent

    def finish(self):
        sys.stdout.write('\n')


def ensure_dir_exists(dest_filename):
    """
    Something we'll need to do often: given a filename we want to write to,
    make sure its containing directory exists.
    """
    base_dir = os.path.dirname(dest_filename)
    if not os.path.exists(base_dir):
        os.makedirs(base_dir)


def download(url, dest_filename):
    """
    Download the file at `url` to `dest_filename`. Show a progress bar
    while downloading.
    """
    ensure_dir_exists(dest_filename)
    tracker = ProgressTracker(url)
    urlretrieve(url, dest_filename, reporthook=tracker.report_progress)
    tracker.finish()
    logger.info("Saved database to %s", dest_filename)


def download_and_extract_raw_data(url=None, root_dir=None):
    """
    Download the .tar.gz of raw data that can be used to build the database.
    """
    if url is None:
        url = config.RAW_DATA_URL

    if root_dir is None:
        root_dir = os.path.dirname(config.RAW_DATA_DIR)

    dest_filename = os.path.join(root_dir, 'wordfreq-data.tar.gz')
    ensure_dir_exists(dest_filename)
    download(url, dest_filename)

    logger.info("Extracting %s" % dest_filename)
    with tarfile.open(dest_filename, 'r') as tarf:
        tarf.extractall(root_dir)


def download_db(url=None, dest_filename=None):
    """
    Download the database itself, so we don't have to build it.
    """
    if url is None:
        url = config.DB_URL

    if dest_filename is None:
        dest_filename = config.DB_FILENAME

    ensure_dir_exists(dest_filename)
    download(url, dest_filename)


def upload_data(upload_path=None):
    """
    Collect the raw data and the database file, and upload them to an
    appropriate directory on the server that hosts downloads.

    This requires that it's running in a reasonable Unix environment,
    and more notably, that it has the proper SSH keys to upload to that
    server.
    """
    if upload_path is None:
        upload_path = config.UPLOAD_PATH
    
    build_dir = tempfile.mkdtemp('.wordfreq')
    version_dir = os.path.join(build_dir, config.MINOR_VERSION)
    os.makedirs(version_dir)

    source_filename = os.path.join(version_dir, 'wordfreq-data.tar.gz')
    logger.info("Creating %s" % source_filename)
    with tarfile.open(source_filename, 'w:gz') as tarf:
        tarf.add(config.RAW_DATA_DIR)

    logger.info("Copying database file %s" % config.DB_FILENAME)
    subprocess.call([
        '/bin/cp',
        config.DB_FILENAME,
        version_dir
    ])

    logger.info("Uploading to %s" % upload_path)
    subprocess.call([
        '/usr/bin/rsync',
        '-avz',
        version_dir,
        upload_path
    ])

    logger.info("Removing build directory %s" % build_dir)
    shutil.rmtree(build_dir)



#!/usr/bin/env python
from setuptools import setup
from distutils.core import Command
from setuptools.command.install import install
from setuptools.command.develop import develop

import os
import sys
import logging
logging.basicConfig(level=logging.INFO)

# Make sure we can import stuff from here.
current_dir = os.path.dirname(__file__)
sys.path.insert(0, current_dir)

from wordfreq import config, transfer

classifiers=[
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Natural Language :: English',
    'Operating System :: MacOS',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Programming Language :: C',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 2.7',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development',
    'Topic :: Text Processing :: Linguistic',]

README_contents = open(os.path.join(current_dir, 'README.txt')).read()
doclines = README_contents.split("\n")


class SimpleCommand(Command):
    """
    Get the boilerplate out of the way for commands that take no options.
    """
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass


class BuildDatabaseCommand(SimpleCommand):
    description = "Build the word frequency database from raw data"
    def run(self):
        from wordfreq.build import load_all_data
        load_all_data()


class DownloadDatabaseCommand(SimpleCommand):
    description = "Download the built word frequency database"
    user_options = []

    def run(self):
        transfer.download_db()


class DownloadRawDataCommand(SimpleCommand):
    description = "Download the raw wordlist data"
    user_options = []

    def run(self):
        transfer.download_and_extract_raw_data()


class UploadDataCommand(SimpleCommand):
    description = "Upload the raw data and database"
    user_options = []

    def run(self):
        transfer.upload_data()


class CustomInstallCommand(install):
    def run(self):
        install.run(self)
        if not os.path.exists(config.DB_FILENAME):
            self.run_command('download_db')


class CustomDevelopCommand(develop):
    def run(self):
        develop.run(self)
        if not os.path.exists(config.DB_FILENAME):
            self.run_command('download_db')


requirements = ['ftfy >= 3']
if sys.version_info.major == 2:
    requirements.append('functools32')

setup(
    name="wordfreq",
    version=config.VERSION,
    maintainer='Luminoso Technologies, Inc.',
    maintainer_email='dev@luminoso.com',
    url='http://github.com/LuminosoInsight/wordfreq/',
    license = "MIT",
    platforms = ["any"],
    description = doclines[0],
    classifiers = classifiers,
    long_description = "\n".join(doclines[2:]),
    packages=['wordfreq'],
    install_requires=requirements,
    cmdclass = {
        'build_db': BuildDatabaseCommand,
        'download_db': DownloadDatabaseCommand,
        'download_raw': DownloadRawDataCommand,
        'upload_data': UploadDataCommand,
        'install': CustomInstallCommand,
        'develop': CustomDevelopCommand
    }
)



import os

# This directory stores versions of the wordfreq database.
DB_DIR = (os.environ.get('WORDFREQ_DATA')
          or os.path.expanduser('~/.cache/wordfreq'))

# When the minor version number increments, the data may change.
VERSION = '0.3.0'
MINOR_VERSION = '.'.join(VERSION.split('.')[:2])

# Put these options together to make a database filename.
DB_FILENAME = os.path.join(DB_DIR, "wordfreq-%s.db" % MINOR_VERSION)

# How many words do we cache the frequencies for?
CACHE_SIZE = 100000

# Where can the data be downloaded from?
DOWNLOAD_URL = (os.environ.get('WORDFREQ_URL')
                or 'http://wordfreq.services.luminoso.com/')
RAW_DATA_URL = os.path.join(DOWNLOAD_URL, MINOR_VERSION, 'wordfreq-data.tar.gz')
DB_URL = os.path.join(DOWNLOAD_URL, MINOR_VERSION,
                      'wordfreq-%s.db' % MINOR_VERSION)

# How do we actually get it there? This is the path, including hostname, to give
# to scp to upload the file.
UPLOAD_PATH = 'baldr:/srv/wordfreq/static/'

# Where should raw data go? Inside the package isn't necessary a good
# place for it, because it might be installed in the system site-packages.
#
# The current directory -- as you're running the setup.py script -- seems
# as reasonable as anything.
RAW_DATA_DIR = './wordfreq_data'



# coding: utf-8
from unicodedata import normalize
from ftfy.fixes import remove_unsafe_private_use


def standardize_word(word):
    u"""
    Apply various normalizations to the text. In languages where this is
    relevant, it will end up in all lowercase letters, with pre-composed
    diacritics.

    Some language-specific gotchas:

    - Words ending with a capital "Σ" in Greek have a lowercase version that
      ends with "ς" on Python 3, but "σ" on Python 2. (Python 3 is
      orthographically correct.) This will lead to different frequencies on
      such Greek words, and different numbers of words in total.

    - Words containing a capital "I" in Turkish will be normalized to a
      lowercase "i", incorrectly, instead of "ı". The effective result is
      that the capitalized versions will not share a word count with the
      lowercase versions.
    """
    return normalize('NFKC', remove_unsafe_private_use(word)).lower()
